#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

// goal: to make a "linked list" of elements in C
typedef struct __node_t node_t;
typedef struct __list_t list_t;
typedef void (*list_iterator_t)(node_t*);

// node_t -- holds elements of list
struct __node_t {
    int key;	
    void* value;
    node_t* next;
};

// list_t -- one of these per list
struct __list_t {
    node_t* head;
};

node_t* allocate_node(int key, void* value, node_t* next) {
    node_t* n = malloc(sizeof(node_t));
    assert(n != NULL);
    n->key = key;
    n->value = value;
    n->next = next;
    return n;
}

// initialize a list - empty list has one DUMMY NODE
// this makes delete and some other funcs simpler
void list_init(list_t* l) {
    l->head = allocate_node(0, NULL, NULL);
}

// free dummy node when done (list must be empty)
void list_fini(list_t* l) {
    assert(l->head->next == NULL);
    free(l->head);
}

// insert: put something into FRONT of list
void list_insert_front(list_t* l, int key, void* value) {
    node_t* n = allocate_node(key, value, l->head->next);
    l->head->next = n;
}

// insert: put something into list in order
void list_insert_ordered(list_t* l, int key, void* value) {
    node_t* n = allocate_node(key, value, NULL);
    node_t *curr, *prev;
    for (curr = l->head->next, prev = l->head; 
	 curr && key > curr->key;
	 prev = curr, curr = curr->next)
	;
    n->next = curr;
    prev->next = n;
}

// insert method #2: insert into BACK of list
void list_insert_back(list_t* l, int key, void* value) {
    node_t *n = allocate_node(key, value, NULL);
    node_t *curr, *prev;
    for (curr = l->head->next, prev = l->head; curr; prev = curr, curr = curr->next)
	;
    prev->next = n;
}

// delete
int list_delete(list_t* l, int key, void** value) {
    node_t *curr, *prev;
    for (curr = l->head->next, prev = l->head; 
	 curr && curr->key != key; 
	 prev = curr, curr = curr->next)
	;
    if (curr == NULL) 
	return -1;
    prev->next = curr->next;
    *value = curr->value;
    free(curr);
    return 0;
}

// lookup
int list_lookup(list_t* l, int key, void** value) {
    node_t *curr;
    for (curr = l->head->next; curr && curr->key != key; curr = curr->next)
	;
    if (curr == NULL) 
	return -1;
    *value = curr->value;
    return 0;
}


// generic list iterator
void list_apply(list_t* l, list_iterator_t func) {
    node_t* curr;
    for (curr = l->head->next; curr; curr = curr->next)
	(*func)(curr);
}


#ifdef TESTING_MAIN

//
// main program
// 

void my_printer(node_t* n) {
    printf("%d value:%d\n", n->key, (int)n->value);
}

#include <stdio.h>
#include <sys/time.h>
#include <assert.h>

double gettime() {
    struct timeval t;
    int rc = gettimeofday(&t, NULL);
    assert(rc == 0);
    return (double)t.tv_sec + (double)t.tv_usec/1e6;
}


int main(int argc, char *argv[]) {
    if (argc != 2) {
	printf("usage: list-6c <num>\n");
	exit(0);
    }

    list_t n;
    list_init(&n);

    double t1 = gettime();

    long long int i;
    for (i = 0; i < atoi(argv[1]); i++) 
	list_insert_front(&n, i, (void*)(i+1));

    double t2 = gettime();

    list_apply(&n, my_printer);

    for (i = 0; i < atoi(argv[1]) * 2; i++) {
	long long v = -1;
	int rc = list_lookup(&n, i, (void *)&v);
	printf("lookup %d %lld\n", rc, v);
    }

    for (i = 0; i < atoi(argv[1]); i++) {
	long long v;
	int rc = list_delete(&n, i, (void *)&v);
	printf("%d %lld\n", rc, v);
    }

    for (i = 0; i < atoi(argv[1]); i++) {
	long long v = -1;
	int rc = list_delete(&n, i, (void *)&v);
	printf("%d %lld\n", rc, v);
    }

    printf("time: %.2f ms\n", 1000 * (t2-t1));
    return 0;
}

#endif
